{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE ExtendedDefaultRules #-}
{-# LANGUAGE QuasiQuotes #-}
{-# OPTIONS_GHC -Wno-unused-local-binds #-}
{-# LANGUAGE TypeApplications #-}
{-# OPTIONS_GHC -Wno-type-defaults #-}
{-# LANGUAGE TupleSections #-}
{-# LANGUAGE RecursiveDo #-}

module Frontend
  ( frontend
  )
where

import Prelude hiding (filter)

import Hydra.Types
import HydraPay.Api as HydraPayApi
import Text.Read (readMaybe)
import Text.Printf (printf)
import System.Random
import qualified Data.Map as Map
import Data.Fixed
import qualified Data.Text as T

import Obelisk.Frontend
import Obelisk.Route
import Obelisk.Route.Frontend
import Reflex.Dom.Core
import Common.Route
import Control.Lens
import Control.Monad.Fix
import Control.Monad.IO.Class
-- import Common.Api as ChannelsApi
import Control.Monad.Trans.Class

import Language.Javascript.JSaddle (jsg, js, liftJSM, fromJSValUnchecked)
import Data.Int (Int64)
import Data.Monoid
import Data.Witherable (catMaybes)
import Data.Maybe (isJust)

default (T.Text)

-- This runs in a monad that can be run on the client or the server.
-- To run code in a pure client or pure server context, use one of the
-- `prerender` functions.
frontend :: Frontend (R FrontendRoute)
frontend = Frontend
  { _frontend_head = do
      el "title" $ text "Hydra Pay Payment Channel Demo"

      elAttr "meta" ("name"=:"viewport" <> "content"=:"width=device-width, initial-scale=1.0, maximum-scale=1.0, user-scalable=no") blank
      elAttr "link" ("rel" =: "preconnect" <> "href" =: "https://fonts.googleapis.com") blank
      elAttr "link" ("rel" =: "preconnect" <> "href" =: "https://fonts.gstatic.com") blank
      elAttr "link" ("href"=:"https://fonts.googleapis.com/css2?family=Inria+Sans:wght@300&family=Inter:wght@100;200;300;400;500;600;700;800;900&family=Krona+One&family=Rajdhani:wght@300;400;500;600;700&display=swap" <> "rel"=:"stylesheet") blank
      elAttr "link" ("rel"=:"stylesheet" <> "href"=:"https://fonts.googleapis.com/css2?family=Material+Symbols+Rounded:opsz,wght,FILL,GRAD@48,400,0,0") blank
      elAttr "script" ("src"=:"https://cdn.tailwindcss.com") blank
  , _frontend_body = elAttr "div" ("class" =: "w-screen h-screen overflow-hidden flex flex-col bg-gray-100" <> "style" =: "font-family: 'Inter', sans-serif;") $ do
      elClass "div" "flex-shrink-0 px-8 py-4 text-xl font-semibold" $ el "h1" $ text "Hydra Pay"
      prerender_ (text "Loading Payment Channel Demo") $ elClass "div" "w-full h-full mt-10 flex-grow px-8 overflow-hidden" $ do
        pb <- getPostBuild
        endpoint <- liftJSM $ do
          let
            webSocketProtocol :: T.Text -> T.Text
            webSocketProtocol "https:" = "wss:"
            webSocketProtocol _ = "ws:"

          wsProtocol <- fmap webSocketProtocol $ fromJSValUnchecked =<< jsg "location" ^. js "protocol"

          theHostname <- fromJSValUnchecked =<< jsg "location" ^. js "hostname"
          let thePort = "8000" -- fromJSValUnchecked =<< jsg "location" ^. js "port"

          pure $ mconcat [ wsProtocol
                       , "//"
                       , theHostname
                       , ":"
                       , thePort
                       , "/hydra/api"
                       ]
        rec
          lastTagId <- foldDyn (+) 0 $ fromIntegral . length <$> sendToHydraPay

          let
            sendToHydraPay =
              attachWith (\tid msgs -> uncurry Tagged <$> zip [tid..] msgs) (current lastTagId) sendMsg

            serverMsg = fmapMaybe id $ rws ^. webSocket_recv

            subscriptions = fmapMaybe (preview _PlainMsg) serverMsg
            responses = fmapMaybe (preview _TaggedMsg) serverMsg

            authResponse = fmapMaybe (preview (_AuthResult) . tagged_payload) responses

          rws :: RawWebSocket t (Maybe ApiMsg) <- jsonWebSocket endpoint $ def
            & webSocketConfig_send .~ sendToHydraPay

          (_, sendMsg :: Event t [ClientMsg]) <- runEventWriterT $ elClass "div" "w-full h-full flex flex-row text-gray-700" $ do
            -- managedDevnetReply <- requester lastTagId responses (GetIsManagedDevnet  <$ pb)
            -- _ <- runWithReplace blank $
            --   mapMaybe ((\case
            --                 IsManagedDevnet x -> Just (monitorView lastTagId responses x)
            --                 _ -> Nothing)
            --              . tagged_payload)
            --   managedDevnetReply
            elClass "div" "flex-grow" $ elClass "div" "max-w-lg" $ mdo
              (_, authenticatedE) <-
                runEventWriterT
                . runWithReplace (tellEvent =<< lift (authentication lastTagId responses))
                . ffor authenticatedE $ \() -> do
                    text "You're authenticated"
                    blank
              runWithReplace blank (fundCreateInit lastTagId responses <$ authenticatedE)
              pure ()
        pure ()
  }


requestPb :: (PostBuild t m, EventWriter t [ClientMsg] m, MonadHold t m) =>
  Dynamic t Int64
  -> Event t (Tagged ServerMsg)
  -> ClientMsg
  -> m (Event t ServerMsg)
requestPb lastId serverMsg msg = do
  pb <- getPostBuild
  requester lastId serverMsg (msg <$ pb)


requestPbExpected :: (PostBuild t m, EventWriter t [ClientMsg] m, MonadHold t m) =>
  Dynamic t Int64
  -> Event t (Tagged ServerMsg)
  -> Getting (Data.Monoid.First b) ServerMsg b
  -> ClientMsg
  -> m (Event t (Maybe b))
requestPbExpected lastId serverMsg sel msg = do
  fmap (preview sel) <$> requestPb lastId serverMsg msg

requestPbExact :: (PostBuild t m, EventWriter t [ClientMsg] m, MonadHold t m) =>
  Dynamic t Int64
  -> Event t (Tagged ServerMsg)
  -> Getting (Data.Monoid.First b) ServerMsg b
  -> ClientMsg
  -> m (Event t b)
requestPbExact lastId serverMsg sel msg = do
  catMaybes <$> requestPbExpected lastId serverMsg sel msg

getAddresses :: (PostBuild t m, EventWriter t [ClientMsg] m, MonadHold t m) =>
  Dynamic t Int64
  -> Event t (Tagged ServerMsg) -> m (Event t [Address])
getAddresses lastId serverMsg =
  requestPbExact lastId serverMsg _DevnetAddresses (GetDevnetAddresses 2)

viewUntil :: _ => Workflow t m (Maybe a) -> m (Event t a)
viewUntil = fmap catMaybes . workflowView

doFundingTx ::
  ( NotReady t m,
    Adjustable t m,
    PostBuild t m,
    EventWriter t [ClientMsg] m,
    MonadHold t m,
    MonadFix m
  ) =>
  Dynamic t Int64 ->
  Event t (Tagged ServerMsg) ->
  Lovelace ->
  TxType ->
  Address ->
  m (Event t Bool)
doFundingTx lastId serverMsg amnt typ addr = viewUntil $ Workflow $ do
  fundsTxE <- requestPbExact lastId serverMsg _FundsTx (GetAddTx typ addr amnt)
  pure . (Nothing,) . ffor fundsTxE $ \tx -> Workflow $ do
    submitSuccessE <- requestPbExpected lastId serverMsg _OperationSuccess (LiveDocEzSubmitTx tx addr)
    pure . (Nothing,) . ffor submitSuccessE $ \v -> Workflow . pure $ (Just (isJust v), never)

-- TODO: Use Either instead of Bool to get the errors/unexpected values reported back
andWhen :: _ => m (Event t Bool) -> m (Event t Bool) -> m (Event t Bool)
andWhen me mc = viewUntil . Workflow $ do
  e <- me
  pure . (Nothing,) . ffor e $ \case
    False -> Workflow $ pure (Just False, never)
    True -> Workflow $ do
      c <- mc
      pure . (Nothing,) . ffor c $ Workflow . pure . (, never) . Just

doAllFunding ::
  ( NotReady t m,
    Adjustable t m,
    PostBuild t m,
    EventWriter t [ClientMsg] m,
    MonadHold t m,
    MonadFix m, _
  ) =>
  Dynamic t Int64 ->
  Event t (Tagged ServerMsg) ->
  Lovelace ->
  [Address] ->
  m (Event t Bool)
doAllFunding lastId serverMsg amnt [one, two] =
  (text "Funding Alice" >> doFundingTx lastId serverMsg amnt Funds one)
  `andWhen` (text "Fueling Alice" >> doFundingTx lastId serverMsg amnt Fuel one)
  `andWhen` (text "Funding Bob" >> doFundingTx lastId serverMsg amnt Funds two)
  `andWhen` (text "Fueling Bob" >> doFundingTx lastId serverMsg amnt Fuel one)

singletonWorkflow :: _ => a -> Workflow t m a
singletonWorkflow = Workflow . pure . (,never)

createAndInitHead :: _ => _ -> _ -> [_] -> m (Event t Bool)
createAndInitHead lastId serverMsg addrs = do
  ((True <$) <$> requestPbExact lastId serverMsg _OperationSuccess (CreateHead (HeadCreate theHeadName addrs)))
    `andWhen`
    ((True <$) <$> requestPbExact lastId serverMsg _OperationSuccess (InitHead (HeadInit theHeadName 3)))




fundCreateInit :: (NotReady t m, Adjustable t m, PostBuild t m,
   EventWriter t [ClientMsg] m, MonadHold t m, MonadFix m, _) =>
  Dynamic t Int64 -> Event t (Tagged ServerMsg) -> m (Event t [Address])
fundCreateInit lastId serverMsg = do
  addrE <- getAddresses lastId serverMsg
  viewUntil . Workflow . pure . (Nothing,) . ffor addrE $ \addrs -> Workflow $ do
    res <- doAllFunding lastId serverMsg (ada 100) addrs
           `andWhen` createAndInitHead lastId serverMsg addrs
    -- TODO: This just assumes all went well, it should check for errors and abort.
    -- (I.e. check whether res == False.)
    pure . (Nothing,) $ Workflow (do
                            text "Head created"
                            pure . (,never) . Just $ addrs)
                         <$ res



        -- gotAddrs <- fmap switchDyn $ prerender (pure never) $ do
        --   rec
        --     jsBuild <- getPostBuild
        --     -- FIXME: something changed here
        --     result <- getAndDecode $ "/demo-addresses" <$ leftmost [jsBuild]
        --                                                            --, () <$ failedToLoad]
        --     let
        --       -- If we failed it is likely the server just hot reloaded in development
        --       -- and we just try again immediately
        --       failedToLoad = fmapMaybe (preview _Nothing) result
        --       addrsRecv = fmapMaybe (preview _Just) result
        --   pure addrsRecv

        -- rec
        --   addrs <- holdDyn (Nothing :: Maybe [Address]) gotAddrs
        --   nextId <- foldDyn (+) (0 :: Int) $ 1 <$ newTx

        --   let
        --     insertNewTx = (current $ Map.insert <$> nextId) <@> newTx

        --   latestPopupTxs <- foldDyn ($) mempty $ mergeWith (.) $ [ insertNewTx
        --                                                          , removeStaleTxs
        --                                                          ]
        --   latestTxs <- foldDyn ($) mempty $ mergeWith (.) $ [ insertNewTx
        --                                                     ]
        --   let
        --     totalTxDisplayTime = 4
        --     -- We assume the first and second index to be Bob and Alice respectively
        --     bobAddress = ffor addrs $ join . fmap (^? ix 0)
        --     aliceAddress = ffor addrs $ join . fmap (^? ix 1)

        --     newTxPopup tx = do
        --       transitionIn <- getPostBuild >>= delay 0.1
        --       transitionOut <- tickLossyFromPostBuildTime totalTxDisplayTime
        --       removeThisPopup <- tickLossyFromPostBuildTime $ totalTxDisplayTime + 2
        --       visible <- holdDyn False $ mergeWith (&&) [True <$ transitionIn, False <$ transitionOut]
        --       let
        --         mkClasses b =
        --           mconcat [ "transition pointer-events-none duration-1000 px-8 py-4 rounded-lg bg-white mb-4 drop-shadow-lg "
        --                   , bool "opacity-0" "opacity-100" b
        --                   ]
        --       elDynClass "div" (mkClasses <$> visible) $ do
        --         elClass "div" "text-xs text-gray-500 mb-2" $ text "New Transaction"

        --         elClass "div" "flex flex-row justify-between items-center" $ do
        --           elClass "div" "flex flex-col mr-16" $ do
        --             elClass "div" "text-gray-600" $ do
        --               text "To "
        --               dynText $ demoTx_to <$> tx
        --             elClass "div" "text-xl" $ do
        --               dynText $ T.pack . printf "%.2f" . lovelaceToAda . demoTx_amount <$> tx
        --               text "ADA"

        --           elClass "div" "flex flex-row text-gray-400" $ do
        --             elClass "div" "flex flex-col" $ do
        --               elClass "div" "text-xs" $ text "Time"
        --               elClass "div" "" $ do
        --                 dynText $ showAsMs . demoTx_time <$> tx
        --                 text "ms"
        --       pure $ () <$ removeThisPopup

        --   -- When we get a server response, go to open a channel
        --   newTx <- appView bobAddress aliceAddress latestTxs

        --   removeStaleTxs <- elClass "div" "pointer-events-none absolute top-0 right-0 p-4 h-full overflow-hidden leading-none" $ elClass "div" "flex flex-col-reverse" $ do
        --     eventList <- listWithKey latestPopupTxs $ \txid tx -> do
        --       removalEv <- newTxPopup tx
        --       pure $ Map.delete txid <$ removalEv

        --     pure $ switchDyn $ mergeWith (.) . Map.elems <$> eventList
        -- subRoute_ $ \case
        --   _ -> do
        --     setRoute $ FrontendRoute_OpenChannel :/ () <$ gotAddrs
        --     setRoute $ FrontendRoute_Setup :/ () <$ postBuild
        -- setRoute $ FrontendRoute_PaymentChannel :/ () <$ newTx
        -- pure ()

data DemoTx = DemoTx
  { demoTx_to :: T.Text
  , demoTx_amount :: Lovelace
  , demoTx_time :: Pico
  }
  deriving (Eq, Show)

showAsMs :: Pico -> T.Text
showAsMs = T.pack . printf "%.2f" . (realToFrac :: Pico -> Float) . (*1000)


-- TODO(skylar): Use Requester from reflex
requester :: (Reflex t, EventWriter t [ClientMsg] m, MonadHold t m) => Dynamic t Int64 -> Event t (Tagged ServerMsg) -> Event t ClientMsg -> m (Event t ServerMsg)
requester lastTagId serverMsg clientMsg = do
  waitingTag <- holdDyn Nothing $ current (Just <$> lastTagId) <@ clientMsg

  let
    isWhatWeAreWaitingFor (Just nid) msg@(Tagged nid' _) | nid == nid' = Just msg
    isWhatWeAreWaitingFor _ _ = Nothing

    properServerMsg = attachWithMaybe isWhatWeAreWaitingFor (current waitingTag) serverMsg

  tellEvent $ pure <$> clientMsg
  pure (tagged_payload <$> properServerMsg)


authentication ::
  ( PostBuild t m
  , SetRoute t (R FrontendRoute) m
  , SetRoute t (R FrontendRoute) m
  , RouteToUrl (R FrontendRoute) m
  , MonadIO (Performable m)
  , PerformEvent t m
  , TriggerEvent t m
  , DomBuilder t m
  , MonadFix m
  , MonadHold t m
  , EventWriter t [ClientMsg] m
  ) => Dynamic t Int64 -> Event t (Tagged ServerMsg) -> m (Event t ())
authentication lastTagId serverMsg = do
    apiKeyInput <- inputElement $ def
                   & initialAttributes .~ ("class" =: "pl-4 pr-2 py-1 bg-transparent text-right" <> "placeholder" =: "Hydra Pay API Key")
    let
      pressedEnter = keypress Enter apiKeyInput

    authReply <- requester lastTagId serverMsg $
      Authenticate <$> current (_inputElement_value apiKeyInput) <@ pressedEnter
    pure . mapMaybe (\case
                 (AuthResult True) -> Just ()
                 _ -> Nothing)
      $ authReply

initialBalances :: _ => _ -> _ -> (_,_) -> m ()
initialBalances lastTagId serverMsg (bobAddress, aliceAddress)= do
      -- Page Title
      elClass "div" "mb-2 font-semibold" $ text "Open a Payment Channel"
      -- Divider
      elClass "div" "mt-2 w-full h-px bg-gray-200" blank

      -- Balances
      elClass "div" "text-2xl mb-2 font-semibold mt-8" $ text "Current Balances"

      elClass "div" "flex flex-row mt-4" $ do
        balanceWidget lastTagId serverMsg "Bob" bobAddress never L1B
        balanceWidget lastTagId serverMsg "Alice" aliceAddress never L1B

      -- Divider
      elClass "div" "mt-2 w-full h-px bg-gray-200" blank

      -- Open Payment Channel UI
      elClass "div" "mt-8" $ do
        elClass "div" "text-sm mb-1" $ text "Amount to Spend"
        _ :: Dynamic t (Maybe Float) <- el "div" $ elClass "div" "w-fit border-2 border-gray-200 flex flex-row items-center" $ do
          amountInput <- inputElement $ def
            & initialAttributes .~ ("class" =: "pl-4 pr-2 py-1 bg-transparent text-right" <> "placeholder" =: "1 ADA" <> "type" =: "number")
            & inputElementConfig_initialValue .~ "1000"
          elClass "span" "mx-2 my-1" $ text "ADA"
          pure $ readMaybe . T.unpack <$> _inputElement_value amountInput
        elClass "div" "text-xs my-1" $ text "100 ADA in fast payment collateral will also be held in hydra pay"
        pure ()

        todo_CONTINUE_WITH_THIS_BUTTON <-  elClass "button" "rounded mt-4 p-4 text-center w-full bg-gray-800 text-white font-bold" $ text "Open Payment Channel"
        pure ()
      pure ()


data BalanceType = L1B | HdB

theHeadName = "channelsHead"

balanceWidget :: _ => _ -> _ -> _ -> _ -> Event t () -> _ -> m (Dynamic t _)
balanceWidget lastTagId serverMsg name addr refetch btype = do
  elClass "div" "text-lg flex flex-col mr-10" $ do
    elClass "div" "font-semibold" $ text name
    -- NOTE(skylar): We assume we have loaded bobAddress if this is visible, so we don't worry about the outer Nothing
    fmap (fmap getFirst . snd) . runDynamicWriterT . elClass "div" "font-semibold" $ mdo
        balanceResult' <- requestPb lastTagId serverMsg ((case btype of
                                                 L1B -> GetL1Balance
                                                 HdB -> GetHeadBalance theHeadName)
                                         addr)

        let balanceResult = fmap (\case
                                 L1Balance b -> Just b
                                 HeadBalance b -> Just b
                                 _ -> Nothing)
                        balanceResult'
        let
          gotBalance = fmapMaybe (preview _Just) balanceResult
          reqFailed = fmapMaybe (preview _Nothing) balanceResult

        delayedFail <- delay 1 reqFailed
        mBalance <- holdDyn (Nothing :: Maybe Float) $ Just . lovelaceToAda <$> gotBalance
        tellDyn (First <$> mBalance)

        dyn_ $ ffor mBalance $ \case
          Nothing -> elClass "div" "animate-pulse bg-gray-700 w-16 h-4" blank
          Just balance -> do
            text $ T.pack . printf "%.2f" $ balance
            text " ADA"

        pure $ maybe 0 id <$> mBalance


paymentChannel :: _ => _ -> _ -> [Address] -> _ -> m (Event t DemoTx)
paymentChannel lastTagId serverMsg [aliceAddress, bobAddress] latestTxs = do
      -- Page Title
      elClass "div" "w-full flex flex-row items-baseline justify-between" $ do
        elClass "div" "mb-2 font-semibold" $ text "Bob & Alice's Payment Channel"
        todo_CONTINUE_WITH_THIS_BUTTON <- elClass "button" "rounded mt-4 px-6 py-2 text-center bg-gray-800 text-white font-semibold" $ text "Close Payment Channel"
        pure ()

      -- Divider
      elClass "div" "mt-2 w-full h-px bg-gray-200" blank

      elClass "div" "text-2xl mb-2 font-semibold mt-8" $ text "L1 Balance"

      elClass "div" "flex flex-row mt-4" $ do
        balanceWidget lastTagId serverMsg "Bob" bobAddress never L1B
        balanceWidget lastTagId serverMsg "Alice" aliceAddress never L1B

      -- Divider
      elClass "div" "mt-2 w-full h-px bg-gray-200" blank

      -- Balances
      elClass "div" "text-2xl mb-2 font-semibold mt-8" $ do
        text "Payment Channel Balance"

      rec
        (bobBalance, aliceBalance) <- elClass "div" "flex flex-row mt-4" $ do
          bb <- balanceWidget lastTagId serverMsg "Bob" bobAddress autoTx HdB
          ab <- balanceWidget lastTagId serverMsg "Alice" aliceAddress autoTx HdB
          pure (bb, ab)

        -- Divider
        elClass "div" "mt-2 w-full h-px bg-gray-200" blank

        -- Statistics
        elClass "div" "text-2xl mb-2 font-semibold mt-8" $ text "Statistics"

        elClass "div" "flex flex-row mt-4" $ do
          elClass "div" "mr-16" $ do
            elClass "div" "text-lg font-semibold" $ text "Total Transactions"
            elClass "div" "text-3xl text-gray-600" $ dynText $ T.pack . show . Map.size <$> latestTxs

          elClass "div" "mr-16" $ do
            elClass "div" "text-lg font-semibold" $ text "Total Time"
            elClass "div" "text-3xl text-gray-600" $ do
              dynText $ showAsMs . Map.foldr ((+) . demoTx_time) 0 <$> latestTxs
              text "ms"

        -- Divider
        elClass "div" "mt-2 w-full h-px bg-gray-200" blank

        -- Action buttons: Send & Automate
        (automateButton, _) <- elClass "div" "" $ do
          todo_CONTINUE_WITH_THIS_BUTTON <- elClass "button" "rounded mt-4 px-6 py-2 text-center bg-gray-800 text-white font-semibold mr-4" $ text "Send ADA"
          elClass' "button" "rounded mt-4 px-6 py-2 text-center bg-gray-800 text-white font-semibold" $ dynText $ ffor automating $ \case
            True -> "Automating"
            False -> "Automate"

        let
          toggleAutomate = domEvent Click automateButton

        automating <- foldDyn ($) False $ not <$ toggleAutomate

        fmap (getFirst . snd) . runEventWriterT . dyn $ ffor automating $ \case
          False -> pure ()
          True -> do
            tick <- tickLossyFromPostBuildTime 1
            randomAmount :: Event t Lovelace <- performEvent $ (liftIO $ ada <$> randomRIO (1,10)) <$ tick
            let
              nextToAddr :: Dynamic t Address = ffor2 bobBalance aliceBalance $ \bb ab ->
                if bb > ab then aliceAddress else bobAddress

              nextFromAddr = ffor2 bobBalance aliceBalance $ \bb ab ->
                if bb > ab then bobAddress else aliceAddress

              bobIsTo = (fmap (== bobAddress) nextToAddr)
              toName = ffor bobIsTo $ \case
                True -> "Bob"
                False -> "Alice"

            sendAdaResponse <- requester  lastTagId serverMsg $
                (\from to am -> SubmitHeadTx from $ HeadSubmitTx theHeadName to am)
                <$> current nextFromAddr
                <*> current nextToAddr
                <@> randomAmount

            -- picoToLatestTx <- holdDyn Nothing $ Just <$> changeDemoBuildTxFunc

            let
              sendSuccess = fmapMaybe (preview _TxConfirmed) $ sendAdaResponse

--            changeDemoBuildTxFunc :: _ = attachWith DemoTx <$> current toName <@> randomAmount

            -- pure $ attachWithMaybe (<*>) (current picoToLatestTx) $ Just <$> sendSuccess
            tellEvent never -- TODO: put DemoTxes here
            pure ()

      -- Transaction History
      elClass "div" "text-2xl mb-2 font-semibold mt-8" $ text "History"

      let
        historyItem tx = do
          elClass "div" "transition pointer-events-none duration-1000 px-4 py-2 rounded bg-white mb-4 drop-shadow " $ do
            elClass "div" "flex flex-row justify-between items-center" $ do
              elClass "div" "flex flex-col mr-16" $ do
                elClass "div" "text-gray-600" $ do
                  text "To "
                  dynText $ demoTx_to <$> tx
                elClass "div" "text-xl" $ do
                  dynText $ T.pack . printf "%.2f" . lovelaceToAda . demoTx_amount <$> tx
                  text "ADA"

              elClass "div" "flex flex-row text-gray-400" $ do
                elClass "div" "flex flex-col" $ do
                  elClass "div" "text-xs" $ text "Time"
                  elClass "div" "" $ do
                    dynText $ showAsMs . demoTx_time <$> tx
                    text "ms"

      let
        noHistory = Map.null <$> latestTxs

      elDynClass "div" "overflow-y-scroll flex-grow" $ dyn_ $ ffor noHistory $ \case
        True -> elClass "div" "text-gray-500" $ text "There is no history yet for this payment channel"
        False -> do
          _ <- elClass "div" "flex flex-col-reverse w-full" $ listWithKey latestTxs $ \_ tx -> historyItem tx
          pure ()

      pure never -- TODO: PUT AUTO TX BACK HERE
  


--    mapMaybe 

    -- dyn_ $ ffor mApiKey $ \case
    --   Nothing -> text "Please enter your API Key"
    --   Just apiKey -> mdo
    --     (authEl, _) <- elClass' "button" "rounded mt-4 p-4 text-center w-full bg-gray-800 text-white font-bold" $ text "Authenticate"

    --     lastTagId <- foldDyn (+) 0 $ fromIntegral . length <$> sendToHydraPay

    --     let
    --       authenticate = Authenticate apiKey <$ domEvent Click authEl

    --       sendMsg =
    --         mergeWith (<>) $ (fmap . fmap) pure $ [authenticate]

    --       sendToHydraPay =
    --         attachWith (\tid msgs -> fmap (uncurry Tagged) $ zip [tid..] msgs) (current lastTagId) sendMsg

    --       serverMsg = fmapMaybe id $ rws ^. webSocket_recv

-- appView :: forall t m.
--   ( PostBuild t m
--   , Prerender t m
--   , SetRoute t (R FrontendRoute) (Client m)
--   , SetRoute t (R FrontendRoute) m
--   , RouteToUrl (R FrontendRoute) m
--   , MonadIO (Performable m)
--   , PerformEvent t m
--   , TriggerEvent t m
--   , DomBuilder t m
--   , MonadFix m
--   , MonadHold t m
--   ) => Dynamic t (Maybe T.Text) -> Dynamic t (Maybe T.Text) -> Dynamic t (Map Int DemoTx) -> RoutedT t (R FrontendRoute) m (Event t DemoTx)
-- appView bobAddress aliceAddress latestTxs = do
--   fmap switchDyn $ elClass "div" "w-full h-full text-gray-700 max-w-4xl mx-auto p-4 rounded flex flex-col" $ subRoute $ \case
--     FrontendRoute_Home -> homeView >> pure never
--     FrontendRoute_Setup -> do
--       elClass "div" "text-3xl flex flex-col justify-center items-center" $ do
--         el "div" $ text "Fast Payments Demo"
--         elClass "div" "text-lg" $ text "populating addresses..."
--         pure never

--     FrontendRoute_OpeningChannel -> do
--       elClass "div" "text-3xl flex flex-col justify-center items-center" $ do
--         el "div" $ text "Fast Payments Demo"

--         let
--           messages = [ "Sending funds from Bob and Alice into Hydra Pay"
--                      , "Creating payment channel"
--                      , "Waiting for payment channel"
--                      ]

--         tick <- tickLossyFromPostBuildTime 2
--         rec
--           currentIndex <- holdDyn (0 :: Int) $ fmap (min (length messages - 1) . (+1)) $ current currentIndex <@ tick
--         elClass "div" "text-lg" $ do
--           dynText $ (messages !!) <$> currentIndex
--           text "..."

--         prerender_ blank $ mdo
--             ws :: RawWebSocket t (Maybe ChannelsApi.ServerMsg) <- lift $ jsonWebSocket "ws://localhost:8000/api" $ (WebSocketConfig @t @ChannelsApi.ClientMsg) wssend never True []
--             let wssend = [DemoInit] <$ _webSocket_open ws
--             setRoute $ FrontendRoute_PaymentChannel :/ () <$ ffilter (== Just InitDone) (_webSocket_recv ws)
--       pure never

--     FrontendRoute_ClosingChannel -> do
--       elClass "div" "text-3xl flex flex-col justify-center items-center" $ do
--         el "div" $ text "Fast Payments Demo"

--         let
--           messages = [ "Closing payment channel"
--                      , "Settling payment channel on L1"
--                      , "Withdrawing funds from payment channel"
--                      ]

--         tick <- tickLossyFromPostBuildTime 5
--         rec
--           currentIndex <- holdDyn (0 :: Int) $ fmap (min (length messages - 1) . (+1)) $ current currentIndex <@ tick
--         elClass "div" "text-lg" $ do
--           dynText $ (messages !!) <$> currentIndex
--           text "..."

--         prerender_ blank $ mdo
--             ws :: RawWebSocket t (Maybe ChannelsApi.ServerMsg) <- lift $ jsonWebSocket "ws://localhost:8000/api" $ (WebSocketConfig @t @ChannelsApi.ClientMsg) wssend never True []
--             let wssend = [CloseFanout] <$ _webSocket_open ws
--             setRoute $ FrontendRoute_OpenChannel :/ () <$ ffilter (== Just CloseFanoutDone) (_webSocket_recv ws)
--       pure never

--     FrontendRoute_OpenChannel -> do
--       -- Page Title
--       elClass "div" "mb-2 font-semibold" $ text "Open a Payment Channel"
--       -- Divider
--       elClass "div" "mt-2 w-full h-px bg-gray-200" blank

--       -- Balances
--       elClass "div" "text-2xl mb-2 font-semibold mt-8" $ text "Current Balances"

--       elClass "div" "flex flex-row mt-4" $ do
--         balanceWidget "Bob" bobAddress never "l1-balance"
--         balanceWidget "Alice" aliceAddress never "l1-balance"

--       -- Divider
--       elClass "div" "mt-2 w-full h-px bg-gray-200" blank

--       -- Open Payment Channel UI
--       elClass "div" "mt-8" $ do
--         elClass "div" "text-sm mb-1" $ text "Amount to Spend"
--         _ :: Dynamic t (Maybe Float) <- el "div" $ elClass "div" "w-fit border-2 border-gray-200 flex flex-row items-center" $ do
--           amountInput <- inputElement $ def
--             & initialAttributes .~ ("class" =: "pl-4 pr-2 py-1 bg-transparent text-right" <> "placeholder" =: "1 ADA" <> "type" =: "number")
--             & inputElementConfig_initialValue .~ "1000"
--           elClass "span" "mx-2 my-1" $ text "ADA"
--           pure $ readMaybe . T.unpack <$> _inputElement_value amountInput
--         elClass "div" "text-xs my-1" $ text "100 ADA in fast payment collateral will also be held in hydra pay"
--         pure ()

--         routeLink (FrontendRoute_OpeningChannel :/ ()) $ elClass "button" "rounded mt-4 p-4 text-center w-full bg-gray-800 text-white font-bold" $ text "Open Payment Channel"
--       pure never

--     FrontendRoute_SendFunds -> mdo
--       -- Page Header
--       elClass "div" "relative w-full flex flex-row items-baseline justify-between" $ do
--         elClass "div" "mb-2 font-semibold" $ text "Bob & Alice's Payment Channel"
--         (backButton, _) <- elClass' "button" "absolute -left-8 rounded w-6 h-6 bg-gray-300 text-center font-semibold" $ text "<"
--         -- NOTE(skylar): Route link prevents the absolute-ness being applied properly, so we just use setRoute!
--         setRoute $ FrontendRoute_PaymentChannel :/ () <$ domEvent Click backButton

--       -- Divider
--       elClass "div" "w-full h-px bg-gray-200" blank

--       -- Balances
--       elClass "div" "text-2xl mb-2 font-semibold mt-8" $ do
--         text "Current Balances "
--         elClass "span" "text-sm" $ text " In payment channel"

--       elClass "div" "flex flex-row mt-4" $ do
--         balanceWidget "Bob" bobAddress never "head-balance"
--         balanceWidget "Alice" aliceAddress never "head-balance"

--       -- Divider
--       elClass "div" "mt-2 w-full h-px bg-gray-200" blank

--       -- Send
--       elClass "div" "text-2xl mb-4 font-semibold mt-8" $ text "Send ADA"

--       -- To/From Selector
--       rec
--         (selectorButton, _) <- elClass' "button" "w-fit bg-gray-300 px-4 py-2 rounded flex gap-6 flex-row" $ do
--           elClass "div" "" $ do
--             elClass "div" "text-left text-xs" $ text "From"
--             elClass "div" "font-semibold" $ dynText $ ffor bobIsTo $ \case
--               True -> "Alice"
--               False -> "Bob"

--           elClass "div" "" $ do
--             elClass "div" "text-left text-xs" $ text "To"
--             elClass "div" "font-semibold" $ dynText $ toName

--         let
--           toName = ffor bobIsTo $ \case
--             True -> "Bob"
--             False -> "Alice"

--           bobIsTo = ((==) <$> toAddr <*> bobAddress)
--           changeSelection = current nextSelection <@ domEvent Click selectorButton
--           fromAddr = ffor3 bobIsTo bobAddress aliceAddress $ \isBob bob alice ->
--             if isBob then alice else bob
--           nextSelection = fromAddr

--         sendBuild <- getPostBuild
--         toAddr <- holdDyn Nothing $ leftmost [current bobAddress <@ sendBuild, changeSelection]
--         pure ()

--       lovelaceSendAmount :: Dynamic t (Maybe Integer) <- el "div" $ elClass "div" "mt-4 w-full border-2 border-gray-200 flex flex-row items-center" $ do
--           amountInput <- inputElement $ def
--             & initialAttributes .~ ("class" =: "text-gray-500 w-full px-8 py-6 bg-transparent text-center text-xl" <> "placeholder" =: "1 ADA" <> "type" =: "number")
--             & inputElementConfig_initialValue .~ "10"
--           elClass "span" "mx-2 my-1" $ text "ADA"
--           pure $ fmap ((round :: Float -> Integer) . ada) . readMaybe . T.unpack <$> _inputElement_value amountInput

--       (sendButton, _) <- elClass' "button" "rounded mt-4 p-4 text-center w-full bg-gray-800 text-white font-bold" $ text "Send ADA"

--       let
--         txSendPayload = liftA2 (HeadSubmitTx "demo") <$> toAddr <*> lovelaceSendAmount
--         sendAda = fmapMaybe (preview _Just) $ current txSendPayload <@ domEvent Click sendButton
--         sendUrl = (fmap ("/submit-tx/"<>) <$> fromAddr)
--         sendReq = liftA2 postJson <$> sendUrl <*> txSendPayload

--         sendSuccess = fmapMaybe (preview _Just) $ (decodeText <=< _xhrResponse_responseText) <$> sendAdaResponse

--         latestTxFunc = liftA2 DemoTx <$> (Just <$> toName) <*> lovelaceSendAmount

--       -- Dynamic function to construct the last transaction
--       picoToLatestTx <- holdDyn Nothing $ attachPromptlyDynWith const latestTxFunc sendAda

--       sendAdaResponse <- fmap switchDyn $ prerender (pure never) $ do
--         performRequestAsync $ fmapMaybe (preview _Just) $ current sendReq <@ sendAda

--       pure $ attachWithMaybe (\mf a -> mf <*> Just a) (current picoToLatestTx) sendSuccess

--     FrontendRoute_PaymentChannel -> do
--       -- Page Title
--       elClass "div" "w-full flex flex-row items-baseline justify-between" $ do
--         elClass "div" "mb-2 font-semibold" $ text "Bob & Alice's Payment Channel"
--         routeLink (FrontendRoute_ClosingChannel :/ ()) $ elClass "button" "rounded mt-4 px-6 py-2 text-center bg-gray-800 text-white font-semibold" $ text "Close Payment Channel"

--       -- Divider
--       elClass "div" "mt-2 w-full h-px bg-gray-200" blank

--       elClass "div" "text-2xl mb-2 font-semibold mt-8" $ text "L1 Balance"

--       elClass "div" "flex flex-row mt-4" $ do
--         balanceWidget "Bob" bobAddress never "l1-balance"
--         balanceWidget "Alice" aliceAddress never "l1-balance"

--       -- Divider
--       elClass "div" "mt-2 w-full h-px bg-gray-200" blank

--       -- Balances
--       elClass "div" "text-2xl mb-2 font-semibold mt-8" $ do
--         text "Payment Channel Balance"

--       rec
--         (bobBalance, aliceBalance) <- elClass "div" "flex flex-row mt-4" $ do
--           bb <- balanceWidget "Bob" bobAddress autoTx "head-balance"
--           ab <- balanceWidget "Alice" aliceAddress autoTx "head-balance"
--           pure (bb, ab)

--         -- Divider
--         elClass "div" "mt-2 w-full h-px bg-gray-200" blank

--         -- Statistics
--         elClass "div" "text-2xl mb-2 font-semibold mt-8" $ text "Statistics"

--         elClass "div" "flex flex-row mt-4" $ do
--           elClass "div" "mr-16" $ do
--             elClass "div" "text-lg font-semibold" $ text "Total Transactions"
--             elClass "div" "text-3xl text-gray-600" $ dynText $ T.pack . show . Map.size <$> latestTxs

--           elClass "div" "mr-16" $ do
--             elClass "div" "text-lg font-semibold" $ text "Total Time"
--             elClass "div" "text-3xl text-gray-600" $ do
--               dynText $ showAsMs . Map.foldr ((+) . demoTx_time) 0 <$> latestTxs
--               text "ms"

--         -- Divider
--         elClass "div" "mt-2 w-full h-px bg-gray-200" blank

--         -- Action buttons: Send & Automate
--         (automateButton, _) <- elClass "div" "" $ do
--           routeLink (FrontendRoute_SendFunds :/ ()) $ elClass "button" "rounded mt-4 px-6 py-2 text-center bg-gray-800 text-white font-semibold mr-4" $ text "Send ADA"
--           elClass' "button" "rounded mt-4 px-6 py-2 text-center bg-gray-800 text-white font-semibold" $ dynText $ ffor automating $ \case
--             True -> "Automating"
--             False -> "Automate"

--         let
--           toggleAutomate = domEvent Click automateButton

--         automating <- foldDyn ($) False $ not <$ toggleAutomate

--         autoTxEv <- dyn $ ffor automating $ \case
--           False -> pure never
--           True -> fmap switchDyn $ prerender (pure never) $ do
--             tick <- tickLossyFromPostBuildTime 1
--             randomAmount :: Event t Lovelace <- performEvent $ (liftIO $ ada <$> randomRIO (1,10)) <$ tick
--             let
--               nextToAddr = join $ ffor2 bobBalance aliceBalance $ \bb ab ->
--                 if bb > ab then aliceAddress else bobAddress

--               nextFromAddr = join $ ffor2 bobBalance aliceBalance $ \bb ab ->
--                 if bb > ab then bobAddress else aliceAddress

--               sendUrl = (fmap ("/submit-tx/"<>) <$> nextFromAddr)

--               bobIsTo = ((==) <$> nextToAddr <*> bobAddress)
--               toName = ffor bobIsTo $ \case
--                 True -> "Bob"
--                 False -> "Alice"

--               changeDemoBuildTxFunc = attachWith DemoTx (current toName) randomAmount

--               submitPayload = attachWith (<*>) (current $ fmap (HeadSubmitTx "demo") <$> nextToAddr) $ Just <$> randomAmount
--               submitReq = attachWithMaybe (<*>) (current $ fmap postJson <$> sendUrl) submitPayload

--             picoToLatestTx <- holdDyn Nothing $ Just <$> changeDemoBuildTxFunc
--             sendAdaResponse <- performRequestAsync submitReq

--             let
--               sendSuccess = fmapMaybe (preview _Just) $ (decodeText <=< _xhrResponse_responseText) <$> sendAdaResponse

--             pure $ attachWithMaybe (<*>) (current picoToLatestTx) $ Just <$> sendSuccess

--         autoTx <- switchHold never autoTxEv

--       -- Transaction History
--       elClass "div" "text-2xl mb-2 font-semibold mt-8" $ text "History"

--       let
--         historyItem tx = do
--           elClass "div" "transition pointer-events-none duration-1000 px-4 py-2 rounded bg-white mb-4 drop-shadow " $ do
--             elClass "div" "flex flex-row justify-between items-center" $ do
--               elClass "div" "flex flex-col mr-16" $ do
--                 elClass "div" "text-gray-600" $ do
--                   text "To "
--                   dynText $ demoTx_to <$> tx
--                 elClass "div" "text-xl" $ do
--                   dynText $ T.pack . printf "%.2f" . lovelaceToAda . demoTx_amount <$> tx
--                   text "ADA"

--               elClass "div" "flex flex-row text-gray-400" $ do
--                 elClass "div" "flex flex-col" $ do
--                   elClass "div" "text-xs" $ text "Time"
--                   elClass "div" "" $ do
--                     dynText $ showAsMs . demoTx_time <$> tx
--                     text "ms"

--       let
--         noHistory = Map.null <$> latestTxs

--       elDynClass "div" "overflow-y-scroll flex-grow" $ dyn_ $ ffor noHistory $ \case
--         True -> elClass "div" "text-gray-500" $ text "There is no history yet for this payment channel"
--         False -> do
--           _ <- elClass "div" "flex flex-col-reverse w-full" $ listWithKey latestTxs $ \_ tx -> historyItem tx
--           pure ()

--       pure autoTx
